/* Notebook control stylesheet */
.lvctrl input { display: inline; }
.lvctrl input[type=range] { width: 200px; border: 0px;}
.lvctrl input[type=button] { width: 100px; border: 1px solid #999; border-radius: 3px; background: transparent; margin-right: 3px; height: 20px;}
.lvctrl input[type=number] { width: 105px; height: 26px; border-radius: 0; border: 1px dotted #999;}
.lvctrl pre { display: inline; }
.lvctrl p  { color: gray; margin: 0px}
.lvctrl label { margin: 0px}

.lvtab    { border: 1px solid #999; padding: 5px; }
.lvtabbar { width:100%; overflow:hidden; margin-top: 5px; }
.lvtabbar .lvtabbar-item { padding:4px 8px; float:left; width:auto; border:none; outline:none; display:block; }
.lvtabbar .dropdown-hover,.lvtabbar .dropdown-click { position:static; float:left; }
.lvtabbar .lvbutton { white-space:normal; }

.lvbutton:hover { color:#000!important; background-color:#ccc!important; border: 1px solid #999; border-bottom: 0; }
.lvbutton { border: 1px solid #999; border-bottom: 0; border-radius: 5px 5px 0px 0px; margin-right: 3px; display:inline-block; outline:0; padding:4px 8px; vertical-align:middle; overflow:hidden; text-decoration:none; color:inherit; background-color:inherit; text-align:center; cursor:pointer; white-space:nowrap; }

.lvseltab,.hover-lvseltab:hover { color:#fff!important; background-color:#555!important; }

/* Range styling: http://danielstern.ca/range.css/?ref=css-tricks#/ */
input[type=range] {
  -webkit-appearance: none;
  width: 100%;
}
input[type=range]:focus {
  outline: none;
}
input[type=range]::-webkit-slider-runnable-track {
  width: 100%;
  height: 3px; /* Larger track in chrome as can only click exactly on it */
  cursor: pointer;
  background: #d7d7d7;
  border-radius: 0px;
  border: 0px solid #7f7f7f;
}
input[type=range]::-webkit-slider-thumb {
  border: 0.5px solid #7f7f7f;
  height: 12px;
  width: 12px;
  border-radius: 8px;
  background: #ffffff;
  cursor: pointer;
  -webkit-appearance: none;
  margin-top: -5px;
}
input[type=range]:focus::-webkit-slider-runnable-track {
  background: #949494;
  margin-top: -3px;
}
input[type=range]::-moz-range-track {
  width: 100%;
  height: 2px;
  cursor: pointer;
  background: #d7d7d7;
  border-radius: 0px;
  border: 0px solid #7f7f7f;
}
input[type=range]::-moz-range-thumb {
  border: 0.5px solid #7f7f7f;
  height: 12px;
  width: 12px;
  border-radius: 8px;
  background: #ffffff;
  cursor: pointer;
}
input[type=range]::-ms-track {
  width: 100%;
  height: 2px;
  cursor: pointer;
  background: transparent;
  border-color: transparent;
  color: transparent;
}
input[type=range]::-ms-fill-lower {
  background: #cacaca;
  border: 0px solid #7f7f7f;
  border-radius: 0px;
}
input[type=range]::-ms-fill-upper {
  background: #d7d7d7;
  border: 0px solid #7f7f7f;
  border-radius: 0px;
}
input[type=range]::-ms-thumb {
  border: 0.5px solid #7f7f7f;
  height: 12px;
  width: 12px;
  border-radius: 8px;
  background: #ffffff;
  cursor: pointer;
  height: 2px;
}
input[type=range]:focus::-ms-fill-lower {
  background: #d7d7d7;
}
input[type=range]:focus::-ms-fill-upper {
  background: #949494;
}

.resizer { display:flex; margin:0; padding:0; resize:both; overflow:hidden }
.resizer > .resized { flex-grow:1; margin:0; padding:0; border:0 }


